function [model alpha cost compt] = llptrain(y, X, S, params)
%
% (Regularized) Logistic label propagation for multi-class classification 
%
% [model alpha cost compt] = llptrain(y, X, S, params)
%
% Input:
%  y - Class labels      [n x 1]
%       NOTE: Unlabeled samples are assigned with NaN.
%  X - Sample vectors    [n x dim]
%  S - Similarity matrix [n x n]
%  params - Parameters        [struct]
%            .C       - Regularization parameter [scalar] (default 1e-2)
%            .maxiter - Maximum number of iterataions [scalar] (default 1000)
%            .eta     - Weight function for known samples [function_handle] (default @(x) max(sum(x,2),1))
%            .init    - Initialization method ['0'|'lr'|'lp'] (default '0')
%                          Numeric array is acceptable to directly give initial values.
%            .verbose - Verbosity [0|1] (default 0)
%
% Output:
%  model  - Classifier model
%            .W - Coefficients for logistic regression [dim+1 x #class-1]
%            .labels - Class labels for the columns [1 x #class]
%  alpha  - Estimated posterior probabilities [n x #class]
%  cost   - Cost value       [scalar]
%  compt  - Computation time [scalar]
%
%
% Reference:
%  T. Kobayashi and N. Otsu,
%  "Logistic Label Propagation,"
%  Pattern Recognition Letters, Vol. 33, No. 5, pp. 580-588, 2012.
%

%- Parameters -%
if ~exist('params','var') || isempty(params), params = struct; end;
params = parseparam(params,{'C',0.01, 'maxiter',1000, 'init','0', 'eta',@(x) max(sum(x,2),1), 'verbose',0});
if params.verbose, params.verbose = 'plot';
else params.verbose = 'off'; end
lambda = params.C;

%- Normalize -%
nstc = struct('mu',mean(X,1)', 'std', std(X,[],1)');
nstc.std = nstc.std + ~nstc.std;
nstc.invstd = 1./nstc.std;
scales = [nstc.invstd.^2; 0]';

[n, dim] = size(X);
dim = dim + 1;

idxu   = isnan(y);
idxk   = ~idxu;
labels = reshape(unique(y(idxk)),1,[]);
ncls   = numel(labels);
alpha  = double(bsxfun(@eq, y, labels));
alphak = alpha(idxk,:);

L = spdiag(sum(S,2)) - S;

Luu = L(idxu,idxu);
Luk_alphak = L(idxu,idxk)*alphak;

eta = params.eta(S(idxk,idxu));

log_ = @(x) log(max(x,eps));

if ischar(params.init)
	switch(upper(params.init))
	case 'LR'
		w0 = llptrain(y(idxk), X(idxk,:), S(idxk,idxk), struct('maxiter',200,'init','0','verbose',false,'lambda',params.lambda,'eta',@(x) eta));
	case 'LP'
		logeps = -log_(0);
		logalpha0 = alpha(:,1:end-1);
		logalpha0(idxk,:) = (2*logalpha0(idxk,:)-1)*logeps;
		logalpha0(idxu,:) = -((spdiag(repmat(1e-4,size(Luu,1),1))+Luu)\Luk_alphak(:,1:end-1));
		logalpha0(idxu,:) = log_( bsxfun(@times, logalpha0(idxu,:), 1./(1-sum(logalpha0(idxu,:),2))) );
		mlogalpha0 = mean(logalpha0,1);
		y_   = X\logalpha0;
		one_ = X\ones(n,1);
		b0 = (nstc.mu'*y_ - mlogalpha0)/(1-nstc.mu'*one_);
		w0 = y_ + one_*b0;
		w0 = [bsxfun(@times, nstc.std, w0); nstc.mu'*w0-b0];
	case '0'
		w0 = zeros(dim, ncls-1);
	otherwise
		error('llptrain:wrongInitmethod','Wrong init: %s', params.init);
	end
else
	%- numeric values -%
	w0 = params.init;
	if size(w0,1) ~= dim || size(w0,2) ~= ncls-1
		error('llptrain:wrongInitsize','Wrong size of initial weights.');
	end
end

tic;
opts   = struct('MaxIter', params.maxiter, 'TolFun', 1e-6, 'TolGrad', 1e-6, 'Convex', 0, 'Display', params.verbose);
mvfunc = @(x) nmv(X,nstc,x);
[w cost] = lbfgs_logit(@objgradfunc,reshape(w0,1,[]),mvfunc,opts);
compt = toc;
w = reshape(w, dim, ncls-1);

bias = w(end,:);
w = bsxfun(@times, w(1:end-1,:), nstc.invstd);
w(end+1,:) = bias-nstc.mu'*w;

if ~strcmp(params.verbose,'off')
	fprintf('cost=%f, |w|^2=%f at %fsec\n', cost, sum(sum(w(1:end-1,:).^2)), compt);
end

model = struct('W',w, 'labels',labels);
[~, alpha(idxu,:)] = llppredict(X(idxu,:),model);

function [cost grad] = objgradfunc(w, Xw)
	alpha_  = logit(Xw); %[n x ncls]
	alphau_ = alpha_(idxu,:);
	alphak_ = alpha_(idxk,:);
	l       = Luu*alphau_+Luk_alphak;
	cost = 0.5*sum(sum(alphau_.*(l+Luk_alphak)) - sum(bsxfun(@times,eta,alphak).*log(alphak_))) + 0.5*lambda*sum(scales*reshape(w.^2, [], ncls-1));

	if nargout > 1
		alphal   = alphau_.*l;
		sumalphal= sum(alphal,2);
		dev = zeros(n, ncls);
		dev(idxu,:) = alphal - bsxfun(@times,sumalphal,alphau_);
		dev(idxk,:) = 0.5*bsxfun(@times, eta, alphak_ - alphak);
		sumdev = sum(dev,1);
		grad = reshape([bsxfun(@times, nstc.invstd, X'*dev(:,1:end-1)-nstc.mu*sumdev(1:end-1));sumdev(1:end-1)],1,[]) + repmat(lambda*scales,1,ncls-1).*w;
	end
end

end

function z = nmv(X, nstc, y)
% 
%  z - X*y
%
Y = reshape(y,size(X,2)+1,[]);
W = bsxfun(@times, nstc.invstd, Y(1:end-1,:));
B = Y(end,:) - nstc.mu'*W;
%%YOU CAN REPLACE '*' by FASTER METHOD%%
z = bsxfun(@plus, X*W, B);
end
